==> great_expectations-1.0.0a3.tar.gz: great_expectations-1.0.0a3/great_expectations/render/renderer/slack_renderer.py <==
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.checkpoint.checkpoint import CheckpointResult
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )


class SlackRenderer(Renderer):
    def render(  # noqa: PLR0913
        self,
        validation_result: ExpectationSuiteValidationResult,
        data_docs_pages: list[dict] | None = None,
        notify_with: list[str] | None = None,
        show_failed_expectations: bool = False,
        validation_result_urls: list[str] | None = None,
    ) -> list[dict]:
        data_docs_pages = data_docs_pages or []
        notify_with = notify_with or []
        validation_result_urls = validation_result_urls or []

        blocks: list[dict] = []

        title_block = self._build_title_block(
            validation_result=validation_result,
            show_failed_expectations=show_failed_expectations,
            validation_result_urls=validation_result_urls,
        )
        blocks.append(title_block)

        report_element_block = self._build_report_element_block(
            data_docs_pages=data_docs_pages, notify_with=notify_with
        )
        if report_element_block:
            blocks.append(report_element_block)

        return blocks

    def _build_title_block(
        self,
        validation_result: ExpectationSuiteValidationResult,
        show_failed_expectations: bool,
        validation_result_urls: list[str],
    ) -> dict:
        expectation_suite_name = validation_result.suite_name
        data_asset_name = validation_result.asset_name or "__no_data_asset_name__"
        n_checks_succeeded = validation_result.statistics["successful_expectations"]
        n_checks = validation_result.statistics["evaluated_expectations"]
        run_id = validation_result.meta.get("run_id", "__no_run_id__")
        batch_id = validation_result.batch_id
        check_details_text = f"*{n_checks_succeeded}* of *{n_checks}* expectations were met"

        failed_expectations_text = ""

        status = "Failed :x:"
        if validation_result.success:
            status = "Success :tada:"
        else:  # noqa: PLR5501
            if show_failed_expectations:
                failed_expectations_text = self.create_failed_expectations_text(
                    validation_result["results"]
                )

        summary_text = ""
        if validation_result_urls:
            # This adds hyperlinks for defined URL
            if len(validation_result_urls) == 1:
                title_hlink = f"*<{validation_result_urls[0]} | Validation Result>*"
            else:
                title_hlink = "*Validation Result*"
            batch_validation_status_hlinks = "".join(
                f"*Batch Validation Status*: *<{validation_result_url} | {status}>*"
                for validation_result_url in validation_result_urls
            )
            summary_text += f"""{title_hlink}
{batch_validation_status_hlinks}
            """
        else:
            summary_text += f"*Batch Validation Status*: {status}"

        summary_text += f"""
*Expectation Suite name*: `{expectation_suite_name}`
*Data Asset Name*: `{data_asset_name}`
*Run ID*: `{run_id}`
*Batch ID*: `{batch_id}`
*Summary*: {check_details_text}"""

        if failed_expectations_text:
            summary_text += failed_expectations_text

        return {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": summary_text,
            },
        }

    def _build_report_element_block(
        self, data_docs_pages: list[dict], notify_with: list[str]
    ) -> dict | None:
        if not data_docs_pages:
            return None

        if notify_with:
            for docs_link_key in notify_with:
                if docs_link_key in data_docs_pages:
                    docs_link = data_docs_pages[docs_link_key]
                    report_element = self._get_report_element(docs_link)
                else:
                    logger.critical(
                        f"*ERROR*: Slack is trying to provide a link to the following DataDocs: `"
                        f"{docs_link_key!s}`, but it is not configured under `data_docs_sites` in the "  # noqa: E501
                        f"`great_expectations.yml`\n"
                    )
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*ERROR*: Slack is trying to provide a link to the following DataDocs: "  # noqa: E501
                            f"`{docs_link_key!s}`, but it is not configured under "
                            f"`data_docs_sites` in the `great_expectations.yml`\n",
                        },
                    }
                if report_element:
                    return report_element
        else:
            for docs_link_key in data_docs_pages:
                if docs_link_key == "class":
                    continue
                docs_link = data_docs_pages[docs_link_key]
                report_element = self._get_report_element(docs_link)
                return report_element

        return None

    def _build_divider_block(self) -> dict:
        return {"type": "divider"}

    def concatenate_text_blocks(
        self, checkpoint_result: CheckpointResult, text_blocks: list[dict]
    ) -> dict:
        checkpoint_name = checkpoint_result.checkpoint_config.name
        status = checkpoint_result.success or False

        documentation_url = "https://docs.greatexpectations.io/docs/terms/data_docs"
        footer_section = {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": f"Learn how to review validation results in Data Docs: {documentation_url}",  # noqa: E501
                }
            ],
        }

        all_blocks = text_blocks + [footer_section]
        return {"blocks": all_blocks, "text": f"{checkpoint_name}: {status}"}

    def _get_report_element(self, docs_link):
        report_element = None
        if docs_link:
            try:
                if "file://" in docs_link:
                    # handle special case since Slack does not render these links
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: `{docs_link}` \n (Please copy and paste link into "  # noqa: E501
                            f"a browser to view)\n",
                        },
                    }
                else:
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: <{docs_link}|{docs_link}>",
                        },
                    }
            except Exception as e:
                logger.warning(
                    f"""SlackRenderer had a problem with generating the docs link.
                    link used to generate the docs link is: {docs_link} and is of type: {type(docs_link)}.
                    Error: {e}"""  # noqa: E501
                )
                return
        else:
            logger.warning("No docs link found. Skipping data docs link in Slack message.")
        return report_element

    def create_failed_expectations_text(self, validation_results: list[dict]) -> str:
        failed_expectations_str = "\n*Failed Expectations*:\n"
        for expectation in validation_results:
            if not expectation["success"]:
                expectation_name = expectation["expectation_config"]["expectation_type"]
                expectation_kwargs = expectation["expectation_config"]["kwargs"]
                failed_expectations_str += self.create_failed_expectation_text(
                    expectation_kwargs, expectation_name
                )
        return failed_expectations_str

    def create_failed_expectation_text(self, expectation_kwargs, expectation_name) -> str:
        expectation_entity = self.get_failed_expectation_domain(
            expectation_name, expectation_kwargs
        )
        if expectation_entity:
            return f":x:{expectation_name} ({expectation_entity})\n"
        return f":x:{expectation_name}\n"

    @staticmethod
    def get_failed_expectation_domain(expectation_name, expectation_config_kwargs: dict) -> str:
        if "expect_table_" in expectation_name:
            return "Table"

        column_name, column_a, column_b, column_list = (
            expectation_config_kwargs.get("column"),
            expectation_config_kwargs.get("column_A"),
            expectation_config_kwargs.get("column_B"),
            expectation_config_kwargs.get("column_list"),
        )
        if column_name:
            return column_name
        elif column_a and column_b:
            return f"{column_a}, {column_b}"
        elif column_list:
            return str(column_list)

==> great_expectations-1.0.3.tar.gz: great_expectations-1.0.3/great_expectations/render/renderer/slack_renderer.py <==
from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)

from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.core import RunIdentifier
    from great_expectations.core.expectation_validation_result import (
        ExpectationSuiteValidationResult,
    )


class SlackRenderer(Renderer):
    def render(
        self,
        validation_result: ExpectationSuiteValidationResult,
        data_docs_pages: dict | None = None,
        notify_with: list[str] | None = None,
        validation_result_urls: list[str] | None = None,
    ) -> list[dict]:
        data_docs_pages = data_docs_pages or {}
        notify_with = notify_with or []
        validation_result_urls = validation_result_urls or []
        blocks: list[dict] = []

        description_block = self._build_description_block(
            validation_result=validation_result,
            validation_result_urls=validation_result_urls,
        )
        blocks.append(description_block)

        report_element_block = self._build_report_element_block(
            data_docs_pages=data_docs_pages, notify_with=notify_with
        )
        if report_element_block:
            blocks.append(report_element_block)

        return blocks

    def _build_description_block(
        self,
        validation_result: ExpectationSuiteValidationResult,
        validation_result_urls: list[str],
    ) -> dict:
        status = "Failed :x:"
        if validation_result.success:
            status = "Success :tada:"

        validation_link = None
        summary_text = ""
        if validation_result_urls:
            if len(validation_result_urls) == 1:
                validation_link = validation_result_urls[0]
            else:
                title_hlink = "*Validation Results*"
                batch_validation_status_hlinks = "".join(
                    f"*<{validation_result_url} | {status}>*"
                    for validation_result_url in validation_result_urls
                )
                summary_text += f"""{title_hlink}
    {batch_validation_status_hlinks}
                """

        expectation_suite_name = validation_result.suite_name
        data_asset_name = validation_result.asset_name or "__no_data_asset_name__"
        summary_text += f"*Asset*: {data_asset_name}  "
        # Slack does not allow links to local files due to security risks
        # DataDocs links will be added in a block after this summary text when applicable
        if validation_link and "file://" not in validation_link:
            summary_text += (
                f"*Expectation Suite*: {expectation_suite_name}  <{validation_link}|View Results>"
            )
        else:
            summary_text += f"*Expectation Suite*: {expectation_suite_name}"

        return {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": summary_text,
            },
        }

    def concatenate_text_blocks(
        self,
        action_name: str,
        text_blocks: list[dict],
        success: bool,
        checkpoint_name: str,
        run_id: RunIdentifier,
    ) -> dict:
        all_blocks = [
            self._build_header(name=action_name, success=success, checkpoint_name=checkpoint_name)
        ]
        all_blocks.append(self._build_run_time_block(run_id=run_id))
        for block in text_blocks:
            all_blocks.append(block)
        all_blocks.append(self._build_divider())

        return {"blocks": all_blocks}

    def _build_header(self, name: str, success: bool, checkpoint_name: str) -> dict:
        status = "Success :white_check_mark:" if success else "Failure :no_entry:"
        return {
            "type": "header",
            "text": {"type": "plain_text", "text": f"{name} - {checkpoint_name} - {status}"},
        }

    def _build_run_time_block(self, run_id: RunIdentifier) -> dict:
        if run_id is not None:
            run_time = datetime.fromisoformat(str(run_id.run_time))
            formatted_run_time = run_time.strftime("%Y/%m/%d %I:%M %p")
        return {
            "type": "section",
            "text": {"type": "plain_text", "text": f"Runtime: {formatted_run_time}"},
        }

    def _build_divider(self) -> dict:
        return {"type": "divider"}

    def _build_footer(self) -> dict:
        documentation_url = "https://docs.greatexpectations.io/docs/terms/data_docs"
        return {
            "type": "context",
            "elements": [
                {
                    "type": "mrkdwn",
                    "text": f"Learn how to review validation results in Data Docs: {documentation_url}",  # noqa: E501
                }
            ],
        }

    def _get_report_element(self, docs_link):
        report_element = None
        if docs_link:
            try:
                # Slack does not allow links to local files due to security risks
                if "file://" in docs_link:
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: `{docs_link}` \n (Please copy and paste link into "  # noqa: E501
                            f"a browser to view)\n",
                        },
                    }
                else:
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*DataDocs* can be found here: <{docs_link}|{docs_link}>",
                        },
                    }
            except Exception as e:
                logger.warning(
                    f"""SlackRenderer had a problem with generating the docs link.
                    link used to generate the docs link is: {docs_link} and is of type: {type(docs_link)}.
                    Error: {e}"""  # noqa: E501
                )
                return
        else:
            logger.warning("No docs link found. Skipping data docs link in Slack message.")
        return report_element

    def _build_report_element_block(
        self, data_docs_pages: dict, notify_with: list[str]
    ) -> dict | None:
        if not data_docs_pages:
            return None

        if notify_with:
            for docs_link_key in notify_with:
                if docs_link_key in data_docs_pages:
                    docs_link = data_docs_pages[docs_link_key]
                    report_element = self._get_report_element(docs_link)
                else:
                    logger.critical(
                        f"*ERROR*: Slack is trying to provide a link to the following DataDocs: `"
                        f"{docs_link_key!s}`, but it is not configured under `data_docs_sites` in the "  # noqa: E501
                        f"`great_expectations.yml`\n"
                    )
                    report_element = {
                        "type": "section",
                        "text": {
                            "type": "mrkdwn",
                            "text": f"*ERROR*: Slack is trying to provide a link to the following DataDocs: "  # noqa: E501
                            f"`{docs_link_key!s}`, but it is not configured under "
                            f"`data_docs_sites` in the `great_expectations.yml`\n",
                        },
                    }
                if report_element:
                    return report_element
        else:
            for docs_link_key in data_docs_pages:
                if docs_link_key == "class":
                    continue
                docs_link = data_docs_pages[docs_link_key]
                report_element = self._get_report_element(docs_link)
                return report_element

        return None
